package com.chelvc.framework.database.support;

import java.util.Collections;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import com.baomidou.mybatisplus.core.metadata.TableInfo;
import com.baomidou.mybatisplus.core.metadata.TableInfoHelper;
import com.baomidou.mybatisplus.core.toolkit.LambdaUtils;
import com.baomidou.mybatisplus.core.toolkit.support.ColumnCache;
import com.baomidou.mybatisplus.core.toolkit.support.SFunction;
import com.baomidou.mybatisplus.core.toolkit.support.SerializedLambda;
import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.chelvc.framework.database.entity.Creatable;
import com.chelvc.framework.database.entity.Deletable;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import lombok.NonNull;
import org.apache.ibatis.reflection.property.PropertyNamer;

/**
 * 数据修改处理工具类
 *
 * @author Woody
 * @date 2024/5/18
 */
public final class Updates {
    private Updates() {
    }

    /**
     * 获取数据更新对象
     *
     * @param <T> 数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> update() {
        return new Update<>();
    }

    /**
     * 更新所有字段
     *
     * @param clazz 对象类型
     * @param <T>   数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> all(@NonNull Class<T> clazz) {
        Update<T> update = update();
        return update.all(clazz);
    }

    /**
     * 更新除指定字段外的所有字段
     *
     * @param clazz   对象类型
     * @param columns 排除字段数组
     * @param <T>     数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> not(@NonNull Class<T> clazz, @NonNull String... columns) {
        Update<T> update = update();
        return update.not(clazz, columns);
    }

    /**
     * 更新除指定字段外的所有字段
     *
     * @param clazz  对象类型
     * @param getter 排除属性Getter方法
     * @param <T>    数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> not(@NonNull Class<T> clazz, @NonNull SFunction<T, ?> getter) {
        Update<T> update = update();
        return update.not(clazz, getter);
    }

    /**
     * 相等
     *
     * @param columns 列名数组
     * @param <T>     数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> eq(@NonNull String... columns) {
        Update<T> update = update();
        return update.eq(columns);
    }

    /**
     * 相等
     *
     * @param getter 属性Getter方法
     * @param <T>    数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> eq(@NonNull SFunction<T, ?> getter) {
        Update<T> update = update();
        return update.eq(getter);
    }

    /**
     * 相加
     *
     * @param columns 列名数组
     * @param <T>     数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> add(@NonNull String... columns) {
        Update<T> update = update();
        return update.add(columns);
    }

    /**
     * 相加
     *
     * @param getter 属性Getter方法
     * @param <T>    数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> add(@NonNull SFunction<T, ?> getter) {
        Update<T> update = update();
        return update.add(getter);
    }

    /**
     * 清空
     *
     * @param columns 列名数组
     * @param <T>     数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> clear(@NonNull String... columns) {
        Update<T> update = update();
        return update.clear(columns);
    }

    /**
     * 清空
     *
     * @param getter 属性Getter方法
     * @param <T>    数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> clear(@NonNull SFunction<T, ?> getter) {
        Update<T> update = update();
        return update.clear(getter);
    }

    /**
     * 相减
     *
     * @param columns 列名数组
     * @param <T>     数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> subtract(@NonNull String... columns) {
        Update<T> update = update();
        return update.subtract(columns);
    }

    /**
     * 相减
     *
     * @param getter 属性Getter方法
     * @param <T>    数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> subtract(@NonNull SFunction<T, ?> getter) {
        Update<T> update = update();
        return update.subtract(getter);
    }

    /**
     * 相乘
     *
     * @param columns 列名数组
     * @param <T>     数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> multiply(@NonNull String... columns) {
        Update<T> update = update();
        return update.multiply(columns);
    }

    /**
     * 相乘
     *
     * @param getter 属性Getter方法
     * @param <T>    数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> multiply(@NonNull SFunction<T, ?> getter) {
        Update<T> update = update();
        return update.multiply(getter);
    }

    /**
     * 相除
     *
     * @param columns 列名数组
     * @param <T>     数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> divide(@NonNull String... columns) {
        Update<T> update = update();
        return update.divide(columns);
    }

    /**
     * 相除
     *
     * @param getter 属性Getter方法
     * @param <T>    数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> divide(@NonNull SFunction<T, ?> getter) {
        Update<T> update = update();
        return update.divide(getter);
    }

    /**
     * 添加更新表达式
     *
     * @param column 列名
     * @param value  值
     * @param <T>    数据类型
     * @return 更新对象实例
     */
    public static <T> Update<T> expression(@NonNull String column, @NonNull String value) {
        Update<T> update = update();
        return update.expression(column, value);
    }

    /**
     * 重复数据更新对象
     *
     * @param <T> 数据类型
     */
    public static class Update<T> {
        private Map<String, ColumnCache> columns;
        private final Map<String, String> expressions = Maps.newHashMapWithExpectedSize(3);

        /**
         * 初始化列缓存
         *
         * @param clazz 对象类型
         */
        private void initializeColumnCache(Class<?> clazz) {
            if (this.columns == null) {
                this.columns = LambdaUtils.getColumnMap(clazz);
            }
            AssertUtils.nonnull(this.columns, () -> "Column cache not found: " + clazz);
        }

        /**
         * 将属性Getter方法转换成列名称
         *
         * @param getter 属性Getter方法
         * @return 列名称
         */
        private String getter2column(SFunction<T, ?> getter) {
            SerializedLambda lambda = LambdaUtils.resolve(getter);
            this.initializeColumnCache(lambda.getInstantiatedType());
            String property = PropertyNamer.methodToProperty(lambda.getImplMethodName());
            ColumnCache column = this.columns.get(LambdaUtils.formatKey(property));
            return column.getColumn();
        }

        /**
         * 更新所有字段
         *
         * @param clazz 对象类型
         * @return 更新对象实例
         */
        public Update<T> all(@NonNull Class<T> clazz) {
            return this.not(clazz);
        }

        /**
         * 更新除指定字段外的所有字段
         *
         * @param clazz   对象类型
         * @param columns 排除字段数组
         * @return 更新对象实例
         */
        public Update<T> not(@NonNull Class<T> clazz, @NonNull String... columns) {
            this.initializeColumnCache(clazz);
            TableInfo table = TableInfoHelper.getTableInfo(clazz);
            boolean creatable = Creatable.class.isAssignableFrom(clazz);
            boolean deletable = Deletable.class.isAssignableFrom(clazz);
            String key = StringUtils.ifEmpty(table.getKeyProperty(), LambdaUtils::formatKey);
            Set<String> excludes = ObjectUtils.isEmpty(columns) ? Collections.emptySet() : Sets.newHashSet(columns);
            this.columns.forEach((property, column) -> {
                if (!Objects.equals(property, key)
                        && !(creatable && ("CREATOR".equals(property) || "CREATETIME".equals(property)))
                        && !(deletable && "DELETED".equals(property)) && !excludes.contains(column.getColumn())) {
                    this.eq(column.getColumn());
                }
            });
            return this;
        }

        /**
         * 更新除指定字段外的所有字段
         *
         * @param clazz  对象类型
         * @param getter 排除属性Getter方法
         * @return 更新对象实例
         */
        public Update<T> not(@NonNull Class<T> clazz, @NonNull SFunction<T, ?> getter) {
            return this.not(clazz, this.getter2column(getter));
        }

        /**
         * 相等
         *
         * @param columns 列名数组
         * @return 更新对象实例
         */
        public Update<T> eq(@NonNull String... columns) {
            for (String column : columns) {
                this.expression(column, " = VALUES(" + column + ")");
            }
            return this;
        }

        /**
         * 相等
         *
         * @param getter 属性Getter方法
         * @return 更新对象实例
         */
        public Update<T> eq(@NonNull SFunction<T, ?> getter) {
            return this.eq(this.getter2column(getter));
        }

        /**
         * 相加
         *
         * @param columns 列名数组
         * @return 更新对象实例
         */
        public Update<T> add(@NonNull String... columns) {
            for (String column : columns) {
                this.expression(column, " = " + column + " + VALUES(" + column + ")");
            }
            return this;
        }

        /**
         * 相加
         *
         * @param getter 属性Getter方法
         * @return 更新对象实例
         */
        public Update<T> add(@NonNull SFunction<T, ?> getter) {
            return this.add(this.getter2column(getter));
        }

        /**
         * 清空
         *
         * @param columns 列名数组
         * @return 更新对象实例
         */
        public Update<T> clear(@NonNull String... columns) {
            for (String column : columns) {
                this.expression(column, " = null");
            }
            return this;
        }

        /**
         * 清空
         *
         * @param getter 属性Getter方法
         * @return 更新对象实例
         */
        public Update<T> clear(@NonNull SFunction<T, ?> getter) {
            return this.clear(this.getter2column(getter));
        }

        /**
         * 相减
         *
         * @param columns 列名数组
         * @return 更新对象实例
         */
        public Update<T> subtract(@NonNull String... columns) {
            for (String column : columns) {
                this.expression(column, " = " + column + " - VALUES(" + column + ")");
            }
            return this;
        }

        /**
         * 相减
         *
         * @param getter 属性Getter方法
         * @return 更新对象实例
         */
        public Update<T> subtract(@NonNull SFunction<T, ?> getter) {
            return this.subtract(this.getter2column(getter));
        }

        /**
         * 相乘
         *
         * @param columns 列名数组
         * @return 更新对象实例
         */
        public Update<T> multiply(@NonNull String... columns) {
            for (String column : columns) {
                this.expression(column, " = " + column + " * VALUES(" + column + ")");
            }
            return this;
        }

        /**
         * 相乘
         *
         * @param getter 属性Getter方法
         * @return 更新对象实例
         */
        public Update<T> multiply(@NonNull SFunction<T, ?> getter) {
            return this.multiply(this.getter2column(getter));
        }

        /**
         * 相除
         *
         * @param columns 列名数组
         * @return 更新对象实例
         */
        public Update<T> divide(@NonNull String... columns) {
            for (String column : columns) {
                this.expression(column, " = " + column + " / VALUES(" + column + ")");
            }
            return this;
        }

        /**
         * 相除
         *
         * @param getter 属性Getter方法
         * @return 更新对象实例
         */
        public Update<T> divide(@NonNull SFunction<T, ?> getter) {
            return this.divide(this.getter2column(getter));
        }

        /**
         * 添加更新表达式
         *
         * @param column 列名
         * @param value  值
         * @return 更新对象实例
         */
        public Update<T> expression(@NonNull String column, @NonNull String value) {
            this.expressions.put(column, value);
            return this;
        }

        /**
         * 排除更新字段
         *
         * @param columns 列名数字
         * @return 更新对象实例
         */
        public Update<T> exclude(@NonNull String... columns) {
            for (String column : columns) {
                this.expressions.remove(column);
            }
            return this;
        }

        /**
         * 排除更新字段
         *
         * @param getter 属性Getter方法
         * @return 更新对象实例
         */
        public Update<T> exclude(@NonNull SFunction<T, ?> getter) {
            return this.exclude(this.getter2column(getter));
        }

        /**
         * 获取表达式字段/值映射表
         *
         * @return 表达式字段/值映射表
         */
        public Map<String, String> getExpressions() {
            return Collections.unmodifiableMap(this.expressions);
        }

        @Override
        public String toString() {
            return this.expressions.toString();
        }
    }
}
