package com.chelvc.framework.email;

import java.util.Objects;

import ch.qos.logback.classic.spi.ILoggingEvent;
import ch.qos.logback.core.AppenderBase;
import ch.qos.logback.core.Layout;
import com.chelvc.framework.common.util.StringUtils;
import lombok.Getter;
import lombok.Setter;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Component;

/**
 * Logback日志邮件发送器
 *
 * @author Woody
 * @date 2024/5/20
 */
@Getter
@Setter
@Component
public class LogbackEmailAppender extends AppenderBase<ILoggingEvent> implements ApplicationContextAware {
    /**
     * 邮件处理器实例
     */
    private static EmailHandler EMAIL_HANDLER;

    private String to;
    private String subject;
    private boolean asynchronous = true;
    private Layout<ILoggingEvent> layout;

    /**
     * 判断日志邮件发送器是否已准备就绪
     *
     * @return true/false
     */
    protected boolean isReady() {
        return this.layout != null && EMAIL_HANDLER != null && StringUtils.notEmpty(this.to);
    }

    /**
     * 获取邮件处理器
     *
     * @return 邮件处理器
     */
    protected EmailHandler getEmailHandler() {
        return Objects.requireNonNull(EMAIL_HANDLER);
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        EMAIL_HANDLER = applicationContext.getBean(EmailHandler.class);
    }

    @Override
    protected void append(ILoggingEvent event) {
        if (!this.isReady()) {
            return;
        }

        // 构建邮件体
        String type = this.layout.getContentType();
        String content = this.layout.doLayout(event);
        Body body = Body.builder().to(this.to).type(type).subject(this.subject).content(content).build();

        // 异步发送邮件
        EmailHandler handler = this.getEmailHandler();
        if (this.asynchronous) {
            this.context.getScheduledExecutorService().execute(() -> handler.send(body));
        } else {
            handler.send(body);
        }
    }
}
